#pragma once
#include <SFML/Graphics.hpp>
#include <unordered_map>
#include <string>
#include <fstream>
namespace zle
{
	static const std::string ZLE_VERSION = "#1.1";

	class TileMap : public sf::Transformable, public sf::Drawable
	{
		sf::VertexArray arr;
		sf::VertexBuffer buff;
		sf::Vector2u mapSize;
		sf::Vector2u tileSize;
		sf::Vector2u textureTileSize;
		sf::Color overAllColor;
		sf::Texture* texture;
		unsigned int textureGap;
		std::vector<std::vector<unsigned short>> mapData;
		void InternalUpdate()
		{
			//resize buffer
			mapData.resize(mapSize.x);
			for (int i = 0; i < mapSize.x; i++)
				mapData[i].resize(mapSize.y);
			arr.setPrimitiveType(sf::Quads);
			arr.resize(4 * mapSize.x * mapSize.y);
			buff.setPrimitiveType(sf::Quads);
			buff.setUsage(sf::VertexBuffer::Dynamic);

			//setup data
			for (int i = 0; i < mapSize.x; i++)
				for (int j = 0; j < mapSize.y; j++)
				{
					sf::Vector2f position = sf::Vector2f(tileSize.x * i, tileSize.y * j);
					unsigned int v = (j + mapSize.y * i) * 4;

					arr[v].position = position;
					arr[v].color = overAllColor;
					arr[v].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 1].position = position + sf::Vector2f(tileSize.x, 0);
					arr[v + 1].color = overAllColor;
					arr[v + 1].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], 0);

					arr[v + 2].position = position + (sf::Vector2f)tileSize;
					arr[v + 2].color = overAllColor;
					arr[v + 2].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);

					arr[v + 3].position = position + sf::Vector2f(0, tileSize.y);
					arr[v + 3].color = overAllColor;
					arr[v + 3].texCoords = sf::Vector2f(mapData[i][j] * textureTileSize.x + textureGap * mapData[i][j], textureTileSize.y);
				}

			//update
			buff.create(arr.getVertexCount());
			buff.update(&arr[0]);
		}
	public:

		TileMap()
		{
			overAllColor = sf::Color::White;
			mapSize = sf::Vector2u(1, 1);
			tileSize = sf::Vector2u(1, 1);
		}
		~TileMap() {}
		void setTileCount(sf::Vector2u mapSize)
		{
			this->mapSize = mapSize;
			InternalUpdate();
		}
		sf::Vector2u getTileCount() const
		{
			return mapSize;
		}
		void setTexture(sf::Texture* texture)
		{
			this->texture = texture;
		}
		const sf::Texture* getTexture() const
		{
			return texture;
		}
		void setTileSize(sf::Vector2u tileSize)
		{
			this->tileSize = tileSize;
			InternalUpdate();
		}
		sf::Vector2u getTileSize() const
		{
			return tileSize;
		}
		void setTextureTileSize(sf::Vector2u textureTileSize)
		{
			this->textureTileSize = textureTileSize;
			InternalUpdate();
		}
		void setTextureGap(unsigned int gap)
		{
			textureGap = gap;
			InternalUpdate();
		}
		unsigned int getTextureGap() const
		{
			return textureGap;
		}
		sf::Vector2u getTextureTileSize() const
		{
			return textureTileSize;
		}
		void setColor(sf::Color color)
		{
			overAllColor = color;
			for (int i = 0; i < mapSize.x; i++)
				for (int j = 0; j < mapSize.y; j++)
				{
					unsigned int v = (j + mapSize.y * i) * 4;

					arr[v].color = overAllColor;
					arr[v + 1].color = overAllColor;
					arr[v + 2].color = overAllColor;
					arr[v + 3].color = overAllColor;
				}
			buff.update(&arr[0]);
		}
		sf::Color getColor() const
		{
			return overAllColor;
		}
		unsigned short getTileID(sf::Vector2u tile) const
		{
			return mapData[tile.x][tile.y];
		}
		void setTileID(sf::Vector2u tile, unsigned short ID)
		{
			mapData[tile.x][tile.y] = ID;

			unsigned int v = (tile.y + mapSize.y * tile.x) * 4;
			arr[v].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID, 0);
			arr[v + 1].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID + textureTileSize.x, 0);
			arr[v + 2].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID + textureTileSize.x, textureTileSize.y);
			arr[v + 3].texCoords = sf::Vector2f(ID * textureTileSize.x + textureGap * ID, textureTileSize.y);
			buff.update(&arr[0]);
		}
	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			states.transform *= getTransform();
			states.texture = texture;
			if (sf::VertexBuffer::isAvailable())
				target.draw(buff, states);
			else
				target.draw(arr, states);
		}
	};
	class Level : public sf::Drawable, public sf::Transformable
	{
		enum class ObjectType
		{
			None,
			Tilemap,
			Sprite,
			Texture,
			Shader_F,
			Shader_V,
		};

		struct Uniforms
		{
			enum class DataTypes
			{
				Bool,
				Float,
				Int,
				Vec2,
				Vec3,
				Vec4,
				IVec2,
				IVec3,
				IVec4,
			};
			std::vector<DataTypes> types;
			std::vector<std::string> uniformNames;
			unsigned int byteSize;
			static int getTypeSize(DataTypes type)
			{
				switch (type)
				{
				case Uniforms::DataTypes::Bool:
					return 1;
					break;
				case Uniforms::DataTypes::Float:
					return 4;
					break;
				case Uniforms::DataTypes::Int:
					return 4;
					break;
				case Uniforms::DataTypes::Vec2:
					return 8;
					break;
				case Uniforms::DataTypes::Vec3:
					return 12;
					break;
				case Uniforms::DataTypes::Vec4:
					return 16;
					break;
				case Uniforms::DataTypes::IVec2:
					return 8;
					break;
				case Uniforms::DataTypes::IVec3:
					return 12;
					break;
				case Uniforms::DataTypes::IVec4:
					return 16;
					break;
				default:
					break;
				}
				return 0;
			}
			unsigned int getIndex(const std::string& name)
			{
				for (int i = 0; i < uniformNames.size(); i++)
				{
					if (name == uniformNames[i])
					{
						return i;
					}
				}
				return 0;
			}
			DataTypes getUniformType(unsigned int index) const
			{
				return types[index];
			}
			int getOffsetToIndex(int index) const
			{
				if (index < 0)
					return -1;
				int offset = 0;
				for (int i = 0; i < index; i++)
				{
					offset += getTypeSize(types[i]);
				}
				return offset;
			}
			static unsigned char* getUniform(unsigned int offset, std::vector<unsigned char>* data)
			{
				return &((*data)[offset]);
			}
			int getIndexFromName(const std::string& name)
			{
				for (int i = 0; i < uniformNames.size(); i++)
				{
					if (uniformNames[i] == name)
						return i;
				}
				return -1;
			}
			void setUniform(unsigned int index, void* uniformPointer, sf::Shader& shader)
			{
				switch (getUniformType(index))
				{
				case Uniforms::DataTypes::Bool:
					shader.setUniform(uniformNames[index], *((bool*)uniformPointer));
					break;
				case Uniforms::DataTypes::Float:
					shader.setUniform(uniformNames[index], *((float*)uniformPointer));
					break;
				case Uniforms::DataTypes::Int:
					shader.setUniform(uniformNames[index], *((int*)uniformPointer));
					break;
				case Uniforms::DataTypes::Vec2:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Vec2*)uniformPointer));
					break;
				case Uniforms::DataTypes::Vec3:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Vec3*)uniformPointer));
					break;
				case Uniforms::DataTypes::Vec4:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Vec4*)uniformPointer));
					break;
				case Uniforms::DataTypes::IVec2:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Ivec2*)uniformPointer));
					break;
				case Uniforms::DataTypes::IVec3:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Ivec3*)uniformPointer));
					break;
				case Uniforms::DataTypes::IVec4:
					shader.setUniform(uniformNames[index], *((sf::Glsl::Ivec4*)uniformPointer));
					break;
				default:
					break;
				}
			}
		};
		struct RenderStateObject
		{
		private:
			friend class Level;
			friend class Drawable;
			sf::Texture texture;
			std::string textureLocation;
			Uniforms uniforms;
			ObjectType type;
		public:
			RenderStateObject()
			{
				type = ObjectType::None;
			}
			~RenderStateObject()
			{

			}
			const ObjectType& getObjectType() const
			{
				return type;
			}
			const sf::Texture& getTexture() const
			{
				return texture;
			}
			const std::string& getTextureLocation() const
			{
				return textureLocation;
			}
		};
		struct Object : public Drawable
		{
		private:
			friend class Level;
			friend class Drawable;
			sf::Sprite* sprite;
			TileMap* tilemap;
			sf::Texture* usingTexture;
			mutable sf::Shader shader;
			mutable std::vector<unsigned char> uniformFData;
			mutable std::vector<unsigned char> uniformVData;
			ObjectType type;
			sf::Int8 drawingZOrder;
			bool visible = 1;
			sf::BlendMode blendMode;
			RenderStateObject* useVShader;
			RenderStateObject* useFShader;
			sf::Color color;

			bool UseSprite()
			{
				if (tilemap || sprite)
					return 0;
				sprite = new sf::Sprite;
				return 1;
			}
			bool UseTileMap()
			{
				if (tilemap || sprite)
					return 0;
				tilemap = new TileMap;
				return 1;
			}
			void* getPointer() const
			{
				if (tilemap)
					return tilemap;
				if (sprite)
					return sprite;
			}
		public:
			Object()
				: drawingZOrder(0)/*, type(ObjectType::None)*/
			{
				sprite = nullptr;
				tilemap = nullptr;
				useVShader = nullptr;
				useFShader = nullptr;
				blendMode = sf::BlendAlpha;
			}
			~Object()
			{
				if (sprite)
					delete sprite;
				if (tilemap)
					delete tilemap;
			}
			const sf::Int8& getDrawingZLayer() const
			{
				return drawingZOrder;
			}
			const bool& isVisible() const
			{
				return visible;
			}
			const ObjectType& getObjectType() const
			{
				return type;
			}
			const sf::BlendMode& getBlendMode() const
			{
				return blendMode;
			}
			Transformable& getTransformable() const
			{
				if (tilemap)
					return (*tilemap);
				if (sprite)
					return (*sprite);
				sf::err() << "Object does not have a transformable!" << std::endl;
				return (*sprite);
			}
			sf::Sprite& getSprite() const
			{
				if (!sprite)
					sf::err() << "Object not a sprite!" << std::endl;
				return (*sprite);
			}
			TileMap& getTileMap() const
			{
				if (!tilemap)
					sf::err() << "Object not a tilemap!" << std::endl;
				return (*tilemap);
			}
			const sf::Shader& getShader() const
			{
				return shader;
			}
			void setDrawingLayer(sf::Int8 drawingLayer)
			{
				drawingZOrder = drawingLayer;
			}
			void setVisible(bool visible)
			{
				this->visible = visible;
			}
			void setBlendMode(const sf::BlendMode& blendMode)
			{
				this->blendMode = blendMode;
			}
			template<typename T>
			void setUniform(const std::string& name, T value)
			{
				shader.setUniform(name, value);
			}

		private:
			virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
			{
				states.blendMode = blendMode;
				if (sprite)
					target.draw(*sprite, states);
				if (tilemap)
					target.draw(*tilemap, states);
			}
		};
		void AnalyzeShader(RenderStateObject& targetObject)
		{
			if (targetObject.type != ObjectType::Shader_F && targetObject.type != ObjectType::Shader_V)
				return;
			std::ifstream load1;
			load1.open(targetObject.textureLocation);
			if (load1.is_open())
			{
				auto stringToType = [](const std::string& str) {
					if (str == "bool")
						return Uniforms::DataTypes::Bool;
					if (str == "int")
						return Uniforms::DataTypes::Int;
					if (str == "float")
						return Uniforms::DataTypes::Float;
					if (str == "ivec2")
						return Uniforms::DataTypes::IVec2;
					if (str == "ivec3")
						return Uniforms::DataTypes::IVec3;
					if (str == "ivec4")
						return Uniforms::DataTypes::IVec4;
					if (str == "vec2")
						return Uniforms::DataTypes::Vec2;
					if (str == "vec3")
						return Uniforms::DataTypes::Vec3;
					if (str == "vec4")
						return Uniforms::DataTypes::Vec4;
				};
				int byteSize = 0;
				targetObject.uniforms.types.clear();
				targetObject.uniforms.uniformNames.clear();
				while (1)
				{
					std::string file;
					load1 >> file;
					if (file.empty())
						break;
					if (file == "uniform")
					{
						load1 >> file;
						if (file != "sampler2D")
						{
							Uniforms::DataTypes type = stringToType(file);
							byteSize += Uniforms::getTypeSize(type);
							targetObject.uniforms.types.push_back(type);
							load1 >> file;
							if (file[file.size() - 1] == ';')
							{
								file.pop_back();
							}
							targetObject.uniforms.uniformNames.push_back(file);
						}
					}
				}
				targetObject.uniforms.byteSize = byteSize;
				load1.close();
			}
		}
		std::unordered_map <std::string, Object> objects;
		std::unordered_map <std::string, RenderStateObject> texture_objects;
	public:
		Level()
		{

		}
		~Level()
		{

		}
		Object& getLevelObject(const std::string& name)
		{
			return objects.at(name);
		}
		RenderStateObject& getRenderStateObject(const std::string& name)
		{
			return texture_objects.at(name);
		}
		bool loadFromFile(const std::string& fileName, std::string resourceFolder)
		{
			sf::BlendMode blendModes[4] = { sf::BlendMode(sf::BlendMode::One, sf::BlendMode::Zero),
			sf::BlendMode(sf::BlendMode::SrcAlpha, sf::BlendMode::One, sf::BlendMode::Add, sf::BlendMode::One, sf::BlendMode::One, sf::BlendMode::Add),
			sf::BlendMode(sf::BlendMode::SrcAlpha, sf::BlendMode::OneMinusSrcAlpha, sf::BlendMode::Add, sf::BlendMode::One, sf::BlendMode::OneMinusSrcAlpha, sf::BlendMode::Add),
			sf::BlendMode(sf::BlendMode::DstColor, sf::BlendMode::Zero) };

			std::ifstream load1;
			load1.open(fileName, std::ios::binary);
			if (load1.is_open())
			{
				std::string line;
				load1 >> line;
				if (line != ZLE_VERSION)
					return 0;
				load1 >> line;
				int counter;
				int temp;
				float tempF;
				load1 >> counter;
				std::vector<sf::Texture*> textureP;
				std::vector<RenderStateObject*> shaderP;
				textureP.resize(counter);
				shaderP.resize(counter);
				for (int i = 0; i < counter; i++)
				{
					load1 >> temp;

					load1 >> line;
					if (temp == 0)
					{
						texture_objects["None"];
						load1 >> temp;
						continue;
					}
					std::string line2 = line;
					int count = 0;
					while (1)
					{
						if (line2[0] == '\\' || line2[0] == '/')
						{
							line2.erase(0, 1);
							line.erase(0, 1);
							count++;
							if (count == 2)
							{
								if (resourceFolder[resourceFolder.size() - 1] != '\\' && resourceFolder[resourceFolder.size() - 1] != '/')
									resourceFolder.push_back('/');
								line.insert(0, resourceFolder);
								break;
							}
						}
						else
						{
							line2.erase(0, 1);
							line.erase(0, 1);
						}

					}
					texture_objects[line2].type = (ObjectType)temp;
					texture_objects[line2].textureLocation = line;
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].texture.loadFromFile(line);
					load1 >> temp;
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].texture.setSmooth(temp);
					load1 >> temp;
					if (texture_objects[line2].type == ObjectType::Texture)
						texture_objects[line2].texture.setRepeated(temp);
					if (texture_objects[line2].type != ObjectType::Texture)
						AnalyzeShader(texture_objects[line2]);
					textureP[i] = &texture_objects[line2].texture;
					shaderP[i] = &texture_objects[line2];

				}
				load1 >> line;
				load1 >> counter;
				for (int i = 0; i < counter; i++)
				{
					load1 >> line;
					load1 >> temp;
					Object& o = objects[line];
					o.color.r = temp;
					load1 >> temp;
					o.color.g = temp;
					load1 >> temp;
					o.color.b = temp;
					load1 >> temp;
					o.color.a = temp;
					load1 >> temp;
					o.type = (ObjectType)temp;
					switch (o.type)
					{
					case ObjectType::Sprite:
						o.UseSprite();
						break;
					case ObjectType::Tilemap:
						o.UseTileMap();
						break;
					}
					load1 >> temp;
					o.usingTexture = textureP[temp];

					load1 >> temp;
					if (temp != -1)
					{
						o.useFShader = shaderP[temp];

						load1 >> line;
						char temp2 = load1.get();
						o.uniformFData.resize(stoi(line));
						for (int i = 0; i < stoi(line); i++)
						{
							temp2 = load1.get();
							o.uniformFData[i] = temp2;
						}
					}

					load1 >> temp;
					if (temp != -1)
					{
						o.useVShader = shaderP[temp];

						load1 >> line;

						o.uniformVData.resize(stoi(line));

						char temp2 = load1.get();
						for (int i = 0; i < stoi(line); i++)
						{
							temp2 = load1.get();
							o.uniformVData[i] = temp2;
						}
					}
					if (o.useVShader && o.useFShader)
					{
						o.shader.loadFromFile(o.useVShader->textureLocation, o.useFShader->textureLocation);
					}
					else if (o.useFShader && !o.useVShader)
						o.shader.loadFromFile(o.useFShader->textureLocation, sf::Shader::Fragment);
					else if (!o.useFShader && o.useVShader)
						o.shader.loadFromFile(o.useVShader->textureLocation, sf::Shader::Vertex);

					if (o.useFShader)
						for (int j = 0; j < o.useFShader->uniforms.types.size(); j++)
						{
							o.useFShader->uniforms.setUniform(j, &o.uniformFData[o.useFShader->uniforms.getOffsetToIndex(j)], o.shader);
						}
					if (o.useVShader)
						for (int j = 0; j < o.useVShader->uniforms.types.size(); j++)
						{
							o.useVShader->uniforms.setUniform(j, &o.uniformFData[o.useVShader->uniforms.getOffsetToIndex(j)], o.shader);
						}

					load1 >> temp;
					o.visible = temp;
					load1 >> temp;
					o.drawingZOrder = temp;
					load1 >> temp;
					o.blendMode = blendModes[temp];

					sf::Vector2f tempV;
					sf::Transformable& t = o.getTransformable();
					load1 >> tempV.x;
					load1 >> tempV.y;
					t.setPosition(tempV);

					load1 >> tempF;
					t.setRotation(tempF);

					load1 >> tempV.x;
					load1 >> tempV.y;
					t.setScale(tempV);

					load1 >> tempV.x;
					load1 >> tempV.y;
					t.setOrigin(tempV);

					switch (o.type)
					{
					case ObjectType::Sprite:
					{
						sf::Sprite* s = o.sprite;
						s->setColor(o.color);

						sf::IntRect tempI;
						load1 >> tempI.left;
						load1 >> tempI.top;
						load1 >> tempI.width;
						load1 >> tempI.height;
						s->setTextureRect(tempI);

						s->setTexture(*o.usingTexture);
					}
					break;
					case ObjectType::Tilemap:
					{
						TileMap* s = o.tilemap;


						load1 >> temp;
						s->setTextureGap(temp);

						load1 >> tempV.x;
						load1 >> tempV.y;
						s->setTextureTileSize((sf::Vector2u)tempV);
						load1 >> tempV.x;
						load1 >> tempV.y;
						s->setTileCount((sf::Vector2u)tempV);
						load1 >> tempV.x;
						load1 >> tempV.y;
						s->setTileSize((sf::Vector2u)tempV);

						s->setColor(o.color);

						for (int i = 0; i < s->getTileCount().x; i++)
							for (int j = 0; j < s->getTileCount().y; j++)
							{
								load1 >> temp;
								s->setTileID(sf::Vector2u(i, j), temp);
							}
						s->setTexture(o.usingTexture);
					}
					break;
					}

				}
				load1.close();
				return 1;
			}
			return 0;
		}


	private:
		virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const
		{
			states.transform *= getTransform();
			for (int i = INT8_MIN; i < INT8_MAX; i++)
			{
				for (auto& n : objects)
				{
					if (!n.second.visible)
						continue;
					if (n.second.drawingZOrder == i)
					{
						if (n.second.useFShader || n.second.useVShader)
							states.shader = &n.second.shader;
						if (!n.second.useFShader && !n.second.useVShader)
							states.shader = nullptr;
						target.draw(n.second, states);
					}
				}
			}
		}
	};
}